package polyglot.ext.ibex.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.types.*;
import polyglot.ext.ibex.types.*;
import polyglot.ext.ibex.visit.*;
import polyglot.ext.ibex.lr.GLR;
import polyglot.visit.*;
import polyglot.util.*;
import java.util.*;

/**
 * Node representing a semantic action.
 */
public class RuleDecl_c extends Term_c implements RuleDecl {
    Flags flags;
    TypeNode type;
    String name;
    List choices;
    boolean isStart;
    Nonterminal nonterminal;
    MethodInstance mi;

    public RuleDecl_c(Position pos, Flags flags, TypeNode type, String name, List choices, boolean isStart) {
        super(pos);
        this.flags = flags;
        this.type = type;
        this.name = name;
        this.choices = TypedList.copyAndCheck(choices, RhsNode.class, true);
        this.isStart = isStart;
    }

    public Flags flags() { return flags; }
    public RuleDecl flags(Flags flags) {
        RuleDecl_c n = (RuleDecl_c) copy();
        n.flags = flags;
        return n;
    }

    public TypeNode type() { return type; }
    public RuleDecl type(TypeNode type) {
        RuleDecl_c n = (RuleDecl_c) copy();
        n.type = type;
        return n;
    }

    public String name() { return name; }
    public RuleDecl name(String name) {
        RuleDecl_c n = (RuleDecl_c) copy();
        n.name = name;
        return n;
    }

    public boolean isStart() { return isStart; }
    public RuleDecl isStart(boolean isStart) {
        RuleDecl_c n = (RuleDecl_c) copy();
        n.isStart = isStart;
        return n;
    }

    public List /* RhsElement */ choices() { return choices; }
    public RuleDecl choices(List choices) {
        RuleDecl_c n = (RuleDecl_c) copy();
        n.choices = TypedList.copyAndCheck(choices, RhsNode.class, true);
        return n;
    }

    public Nonterminal nonterminal() { return nonterminal; }
    public RuleDecl nonterminal(Nonterminal nonterminal) {
        RuleDecl_c n = (RuleDecl_c) copy();
        n.nonterminal = nonterminal;
        return n;
    }

    public CodeInstance codeInstance() {
        return nonterminal();
    }

    protected RuleDecl_c reconstruct(TypeNode type, List choices) {
        if (type != this.type || ! CollectionUtil.equals(choices, this.choices)) {
            RuleDecl_c n = (RuleDecl_c) copy();
            n.type = type;
            n.choices = TypedList.copyAndCheck(choices, RhsNode.class, true);
            return n;
        }
        return this;
    }

    public Node visitChildren(NodeVisitor v) {
        TypeNode type = (TypeNode) visitChild(this.type, v);
        List choices = visitList(this.choices, v);
        return reconstruct(type, choices);
    }

    public NodeVisitor buildTypesEnter(TypeBuilder tb) throws SemanticException {
        return tb.pushCode();
    }

    public Node buildTypes(TypeBuilder tb) throws SemanticException {
        IbexTypeSystem ts = (IbexTypeSystem) tb.typeSystem();

        List l = new ArrayList(choices.size());
        for (int i = 0; i < choices.size(); i++) {
            RhsNode n = (RhsNode) choices.get(i);
            l.add(n.rhs());
        }

        ParserType ct = (ParserType) tb.currentClass();

        Nonterminal nonterminal = ts.nonterminal(position(), ct,
            Flags.NONE, ts.unknownType(position()), name, l, isStart);

        return nonterminal(nonterminal);
    }

    /** Build type objects for the method. */
    public NodeVisitor disambiguateEnter(AmbiguityRemover ar) throws SemanticException {
        if (ar.kind() == AmbiguityRemover.SUPER) {
            return ar.bypassChildren(this);
        }

        return ar;
    }

    public Node disambiguate(AmbiguityRemover ar) throws SemanticException {
        Context c = ar.context();
        IbexTypeSystem ts = (IbexTypeSystem) ar.typeSystem();

        MutableParserType ct = (MutableParserType) c.currentClassScope();

        if (ar.kind() == AmbiguityRemover.SIGNATURES) {
            Nonterminal nonterminal = makeNonterminal(ct, ts);
            return nonterminal(nonterminal).flags(nonterminal.flags());
        }

        if (ar.kind() == AmbiguityRemover.ALL) {
            Nonterminal template = makeNonterminal(ct, ts);

            Nonterminal nonterminal = ct.nonterminalNamed(name);
            nonterminal.setSymbolType(template.symbolType());
            nonterminal.setChoices(template.choices());

            return nonterminal(nonterminal).flags(nonterminal.flags());
        }

        return this;
    }

    public NodeVisitor addMembersEnter(AddMemberVisitor am) {
        MutableParserType ct = (MutableParserType)
            am.context().currentClassScope();
        ct.addNonterminal(nonterminal);
        ct.addMethod(nonterminal);
        return am.bypassChildren(this);
    }

    public Context enterScope(Context c) {
        c = c.pushCode(nonterminal);
        return c;
    }

    /** Type check the method. */
    public Node typeCheck(TypeChecker tc) throws SemanticException {
        IbexTypeSystem ts = (IbexTypeSystem) tc.typeSystem();

        // Check for duplicate choices
        for (int i = 0; i < nonterminal.choices().size(); i++) {
            Rhs rhs = (Rhs) nonterminal.choices().get(i);

            for (int j = i+1; j < nonterminal.choices().size(); j++) {
                Rhs rhsj= (Rhs) nonterminal.choices().get(j);

                if (rhsj.matches(rhs)) {
                    throw new SemanticException("Duplicate rule: \"" + rhs +
                        "\" and \"" + rhsj + "\".", rhsj.position());
                }
            }
        }

        // Check that each merge choice corresponds to a pair of
        // non-merge choices.
        //
        // FIXME: consider inherited rules too.
        for (Iterator i = nonterminal.choices().iterator(); i.hasNext(); ) {
            Rhs rhs = (Rhs) i.next();

            if (rhs instanceof RhsMerge) {
                RhsCase case1 = ((RhsMerge) rhs).rhs1();
                RhsCase case2 = ((RhsMerge) rhs).rhs2();

                boolean found1 = false;
                boolean found2 = false;

                for (Iterator j = nonterminal.choices().iterator(); j.hasNext(); ) {
                    Rhs rhsj = (Rhs) j.next();
                    if (rhsj instanceof RhsCase) {
                        if (rhsj.matches(case1)) {
                            found1 = true;
                        }
                        if (rhsj.matches(case2)) {
                            found2 = true;
                        }
                    }
                }

                if (! found1) {
                    throw new SemanticException("Could not find matching rule for \"" + case1 + "\".", case1.position());
                }
                if (! found2) {
                    throw new SemanticException("Could not find matching rule for \"" + case2 + "\".", case2.position());
                }
            }
        }

        if (tc.context().currentClass().flags().isInterface()) {
            if (flags().isProtected() || flags().isPrivate()) {
                throw new SemanticException("Interface methods must be public.",
                                            position());
            }
        }

        if (! tc.context().currentClass().isSubtype(ts.runtimeParserType())) {
            throw new SemanticException("A parser rule can only appear in a subtype of \"" + ts.runtimeParserType() + "\".", position());
        }

        if (flags().isAbstract() && ! tc.context().currentClass().flags().isAbstract()) {
            throw new SemanticException("Nonterminal \"" + name() + "\" is abstract, but is a member of non-abstract class \"" + tc.context().currentClass() + "\".",
                position());
        }

        if (flags().isStatic()) {
            throw new SemanticException("Nonterminals cannot be static.", position());
        }
        if (flags().isNative()) {
            throw new SemanticException("Nonterminals cannot be native.", position());
        }
        if (flags().isTransient()) {
            throw new SemanticException("Nonterminals cannot be transient.", position());
        }
        if (flags().isStrictFP()) {
            throw new SemanticException("Nonterminals cannot be strictfp.", position());
        }
        if (flags().isVolatile()) {
            throw new SemanticException("Nonterminals cannot be volatile.", position());
        }

        boolean allHaveBody = true;
        boolean noneHaveBody = true;

        for (Iterator i = this.choices.iterator(); i.hasNext(); ) {
            RhsNode n = (RhsNode) i.next();
            if (n.action() != null) {
                noneHaveBody = false;
            }
            else {
                allHaveBody = false;
            }
        }

        if (! flags().isAbstract() && ! allHaveBody && ! type.type().isVoid()) {
            throw new SemanticException("Missing semantic action.", position());
        }

        if (flags().isAbstract() && ! noneHaveBody) {
            throw new SemanticException("Abstract rules cannot have semantic actions.", position());
        }

        return this;
    }

    protected Nonterminal makeNonterminal(ParserType ct, IbexTypeSystem ts)
        throws SemanticException {

        List choices = new LinkedList();

        for (Iterator i = this.choices.iterator(); i.hasNext(); ) {
            RhsNode n = (RhsNode) i.next();
            choices.add(n.rhs());
        }

        Flags flags = this.flags;

        if (ct.flags().isInterface()) {
            flags = flags.Public().Abstract();
        }

        return ts.nonterminal(position(), ct, flags,
                              type.type(), name, choices, isStart);
    }

    /**
     * Return the first (sub)term performed when evaluating this
     * term.
     */
    public Term entry() {
        return this;
    }

    /**
     * Visit this term in evaluation order.
     */
    public List acceptCFG(CFGBuilder v, List succs) {
        return succs;
    }

    public List rewrite(Rewriter rw) throws SemanticException {
        List l = new ArrayList(choices.size());

        NodeFactory nf = rw.nodeFactory();
        IbexTypeSystem ts = (IbexTypeSystem) rw.typeSystem();
        ParsedClassType ct = rw.context().currentClassScope();

        // Don't add action methods to the interface.
        if (! ct.flags().isInterface()) {
            // Add action methods for each RHS choice.
            for (Iterator i = choices.iterator(); i.hasNext(); ) {
                RhsNode rhs = (RhsNode) i.next();

                List formals = new ArrayList();
                List formalTypes = new ArrayList();
                int index = 0;

                String methodName = rhs.rhs().actionMethodName(nonterminal);

                for (Iterator j = rhs.allElements().iterator(); j.hasNext(); ) {
                    RhsElement e = (RhsElement) j.next();
                    index++;

                    Symbol s = e.symbolInstance();

                    if (e instanceof Tagged) {
                        String name = ((Tagged) e).tag();

                        LocalInstance li =
                            ts.localInstance(e.position(), Flags.FINAL, s.symbolType(), name);

                        Formal f = nf.Formal(e.position(), Flags.FINAL,
                                             nf.CanonicalTypeNode(e.position(),
                                                                  s.symbolType()),
                                             name);
                        f = f.localInstance(li);
                        formals.add(f);
                        formalTypes.add(s.symbolType());
                    }
                }

                MethodInstance mi = ts.methodInstance(
                    rhs.position(), ct, flags, type.type(), methodName,
                    formalTypes, Collections.EMPTY_LIST);

                MethodDecl m =
                    nf.MethodDecl(rhs.position(), flags, type, methodName,
                        formals, Collections.EMPTY_LIST, rhs.action());
                m = m.methodInstance(mi);

                ct.addMethod(mi);

                l.add(m);
            }
        }

        GLR glr = rw.glr();

        if (glr.isStartSymbol(nonterminal)) {
            // Add a method for parsing the nonterminal.
            Block b = null;

            int stateNum = glr.startStateNumber(nonterminal);
            int symbolNum = glr.startSymbolNumber(nonterminal);

            ClassType parser = ts.runtimeGLRParserType();

            TypeNode tn = nf.CanonicalTypeNode(position(), parser);

            List formalTypes = new ArrayList(2);
            formalTypes.add(ts.Int());
            formalTypes.add(ts.Int());

            MethodInstance pmi;
            
            try {
                pmi = ts.findMethod(parser, "parse", formalTypes, ct);
            }
            catch (SemanticException e) {
                throw new InternalCompilerError(e);
            }

            if (! flags.isAbstract()) {
                // { return (T) parse(N); }
                // or, if T = void
                // { parse(N); }

                Expr t = nf.This(position());               
                t = t.type(ct);
                Expr n = nf.New(position(), tn, Collections.singletonList(t));
                n = n.type(parser);

                Expr state = nf.IntLit(position(), IntLit.INT, stateNum);
                state = state.type(ts.Int());
                Expr symbol = nf.IntLit(position(), IntLit.INT, symbolNum);
                symbol = symbol.type(ts.Int());

                List args = new ArrayList(2);
                args.add(state);
                args.add(symbol);

                Call c = nf.Call(position(), n, pmi.name(), args);
                c = c.methodInstance(pmi);
                c = (Call) c.type(pmi.returnType());

                Stmt s;

                if (type.type().isVoid()) {
                    s = nf.Eval(position(), c);
                }
                else {
                    // Cast to the appropriate type.
                    Expr e = nf.Cast(position(), type, c);
                    e = e.type(type.type());
                    s = nf.Return(position(), e);
                }

                b = nf.Block(position(), Collections.singletonList(s));
            }

            /*
            MethodInstance mi = ts.methodInstance(
                position(), ct, flags, type.type(), name(),
                Collections.EMPTY_LIST, pmi.throwTypes());
                */

            List throwTypes = new ArrayList(nonterminal.throwTypes().size());

            for (Iterator i = nonterminal.throwTypes().iterator(); i.hasNext(); ) {
                Type t = (Type) i.next();
                throwTypes.add(nf.CanonicalTypeNode(position(), t));
            }

            MethodDecl m =
                nf.MethodDecl(position(), flags, type, name(),
                    Collections.EMPTY_LIST, throwTypes, b);
            m = m.methodInstance(nonterminal);

            /*
            ct.addMethod(mi);
            */

            l.add(m);
        }

        ClassType mergeable = ts.runtimeMergeableType();

        // Check that there is a merge action, if needed.
        if (true ||
            ! nonterminal.symbolType().isSubtype(mergeable) &&
            ! nonterminal.symbolType().isVoid()) {

            // Get the merge pairs.
            Collection groups = glr.mergeRulesForNonterminal(nonterminal);

            // Check for groups that are not covered by a merge action.
            // Remove from the collection any covered groups.

            for (Iterator i = groups.iterator(); i.hasNext(); ) {
                Collection group = (Collection) i.next();

                boolean found = false;
            
                for (Iterator j = nonterminal.choices().iterator(); j.hasNext(); ) {
                    Rhs rhs = (Rhs) j.next();

                    if (rhs instanceof RhsMerge) {
                        RhsCase case1 = ((RhsMerge) rhs).rhs1();
                        RhsCase case2 = ((RhsMerge) rhs).rhs2();

                        boolean found1 = false;
                        boolean found2 = false;

                        for (Iterator k = group.iterator(); k.hasNext(); ) {
                            Rhs rhsk = (Rhs) k.next();

                            if (rhsk.matches(case1)) {
                                found1 = true;
                            }
                            if (rhsk.matches(case2)) {
                                found2 = true;
                            }
                        }

                        if (found1 && found2) {
                            found = true;
                            break;
                        }
                    }
                }

                if (found) {
                    i.remove();
                }
            }

            // Now check that all groups were covered.
            if (groups.size() > 0) {
                String message = "The following pairs of rules for nonterminal \"" + nonterminal + "\" are ambiguous and require either a merge action, or the semantic action return type \"" + nonterminal.symbolType() + "\" must implement \"" + mergeable + "\" or be void:\n";

                for (Iterator i = groups.iterator(); i.hasNext(); ) {
                    Collection group = (Collection) i.next();

                    message += "\n";

                    for (Iterator j = group.iterator(); j.hasNext(); ) {
                        Rhs rhs = (Rhs) j.next();
                        message += "    " + nonterminal + " ::= " + rhs + "\n";
                    }
                }

                throw new SemanticException(message, position());
            }
        }

        return l;
    }
}
