package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

/**
 * This class contains code that converts a context free grammar into
 * a DFA for a regular approximation of the language.
 *
 * The basic translation is due to Nederhof and is decribed in the paper
 * "Practical Experiments with Regular pproximation" in Computational
 * Linguistics, 26(1).  Section 4.1.
 */
class ApproximationFA
{
    List terminals;
    List nonterminals;
    List allStates;
    NFAState startState;
    NFAState acceptState;
    GLRNonterminal startSymbol;
    List alphabet;

    /** Map from nonterminals to terminals created for the nonterminal. */
    NFATerminal[] ntTerminals;

    /** NFAs for each nonterminal. */
    NFA[] nfas;

    /**
     * Build a regular approximation for the rule's RHS.
     * The LHS must be a augmented grammar start symbol; it cannot
     * be reachable from the RHS.
     */
    ApproximationFA(List terminals, List nonterminals, GLRNonterminal S) {
        this.terminals = new ArrayList();
        this.nonterminals = nonterminals;
        this.startSymbol = S;

        this.allStates = new ArrayList();

        if (S.rules().size() != 1) {
            throw new InternalCompilerError(
                "Cannot create NFA for non-augmented grammar.");
        }

        this.ntTerminals = new NFATerminal[nonterminals.size()];
        this.alphabet = new ArrayList(terminals.size() + nonterminals.size());
        alphabet.addAll(terminals);

        // Create a nonterminal-terminal for each nonterminal.
        for (Iterator i = nonterminals.iterator(); i.hasNext(); ) {
            GLRNonterminal A = (GLRNonterminal) i.next();
            ntTerminal(A);
        }

        this.partition = new Collection[nonterminals.size()];
        this.partitions = new UnionFind();
        this.recursive = new BitSet();
        findPartitions(S);

        // Now for each partition, determine if the partition is left-
        // or right- generating.  Using this information, partitions can
        // be classified as left, right, cyclic, or self recursive.
        this.leftGenerating = new BitSet();
        this.rightGenerating = new BitSet();
        computeLeftRightGeneratingDFS(S, new BitSet());

        System.out.println("partitions for " + S);
        for (int i = 0; i < partition.length; i++) {
            if (partition[i] != null) {
                System.out.print(i + ": " + partition[i] + " ");
                switch (recursive(i)) {
                    case NONE: System.out.println("non-recursive"); break;
                    case LEFT: System.out.println("left"); break;
                    case RIGHT: System.out.println("right"); break;
                    case CYCLIC: System.out.println("cyclic"); break;
                    case SELF: System.out.println("self"); break;
                }
            }
        }

        // Compute partitions of mutually recursive nonterminals.
        // Now for each mutually recursiver partition, generate
        // a DFA for the partition.
        // Nonterminals not in the partition are treated as terminals.

        nfas = new NFA[nonterminals.size()];

        if (recursive(partitions.find(S.index())) != NONE) {
            throw new InternalCompilerError("Cannot have recursive start symbol.");
        }

        NonterminalNFA nfa = new NonterminalNFA(S, alphabet());
        nfas[S.index()] = nfa;

        for (Iterator j = S.rules.iterator(); j.hasNext(); ) {
            GLRRule rule = (GLRRule) j.next();
            makeFA(nfa, nfa.q0, rule.rhs, nfa.q1);
            nfas[S.index()] = new DFA(nfa);
        }

        for (Iterator k = nonterminals.iterator(); k.hasNext(); ) {
            GLRNonterminal A = (GLRNonterminal) k.next();

            if (nfas[A.index()] == null &&
                recursive(partitions.find(A.index())) == NONE) {

                System.out.println("Missed NFA for non-recursive nonterminal " + A);

                NonterminalNFA Anfa = new NonterminalNFA(A, alphabet());
                nfas[A.index()] = Anfa;

                for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                    GLRRule rule = (GLRRule) j.next();
                    makeFA(Anfa, Anfa.q0, rule.rhs, Anfa.q1);
                }

                nfas[A.index()] = new DFA(Anfa);
            }
        }

        for (Iterator k = nonterminals.iterator(); k.hasNext(); ) {
            GLRNonterminal A = (GLRNonterminal) k.next();
            nfas[A.index()].dump();
        }
    }

    public List alphabet() {
        return alphabet;
    }

    class NonterminalNFA extends AbstractNFA {
        NFAState q0, q1;
        GLRNonterminal A;
        List states;

        NonterminalNFA(GLRNonterminal A, List terminals) {
            super(terminals);
            this.states = new ArrayList();
            this.A = A;
            this.q0 = freshState(A.toString() + "0");
            this.q1 = freshState(A.toString() + "1");
        }

        public NFAState start() {
            return q0;
        }

        NFAState freshState(final String s) {
            NFAState q = new NFAState(states.size()) {
                public String toString() { return s; } };
            states.add(q);
            return q;
        }

        NFAState freshState(GLRNonterminal A) {
            NFAState q = new NonterminalState(A, states.size());
            states.add(q);
            return q;
        }

        public boolean accepting(NFAState q0, NFAState q1) {
            return q0 == this.q0 && q1 == this.q1;
        }

        public Iterator nodes() {
            return states.iterator();
        }

        public int size() {
            return states.size();
        }

        public NFAState nodeAt(int i) {
            return (NFAState) states.get(i);
        }

        public NFAEdge edge(NFAState q1, NFATerminal label, NFAState q2) {
            NFAEdge e = new NFAEdge(q1, q2, label);
            return e;
        }

        public String toString() {
            return "approximation NFA for " + A;
        }
    }

    public NFAState nodeAt(int i) {
        return (NFAState) allStates.get(i);
    }

    public Iterator nodes() {
        return allStates.iterator();
    }

    public NFAState start() { return startState; }
    public NFAState accept() { return acceptState; }

    public boolean accepting(NFAState q0, NFAState q) {
        return q0 == start() && q == accept();
    }

    public int size() { return allStates.size(); }

    static class NonterminalState extends NFAState {
        GLRNonterminal A;

        NonterminalState(GLRNonterminal A, int index) {
            super(index);
            this.A = A;
        }

        public String toString() {
            return A.toString();
        }
    }

    static GLRSymbol leftHead(List alpha) {
        if (alpha.size() > 0) {
            return (GLRSymbol) alpha.get(0);
        }

        return null;
    }

    static GLRSymbol rightHead(List alpha) {
        if (alpha.size() > 0) {
            return (GLRSymbol) alpha.get(alpha.size()-1);
        }

        return null;
    }

    static List leftTail(List alpha) {
        if (alpha.size() > 1) {
            return alpha.subList(1,alpha.size());
        }
        return Collections.EMPTY_LIST;
    }

    static List rightTail(List alpha) {
        if (alpha.size() > 1) {
            return alpha.subList(0,alpha.size()-1);
        }
        return Collections.EMPTY_LIST;
    }

    public static final int NONE = 0;
    public static final int LEFT = 1;
    public static final int RIGHT = 2;
    public static final int CYCLIC = 3;
    public static final int SELF = 4;

    BitSet leftGenerating;
    BitSet rightGenerating;
    BitSet recursive;

    UnionFind partitions;
    Collection[] partition;

    void collectPostorderDFS(GLRNonterminal A, List postorder, List[] preds, BitSet visited) {
        if (visited.get(A.index())) {
            return;
        }

        visited.set(A.index());

System.out.println("preorder " + A);
System.out.println("  rules " + A.rules());

        for (Iterator j = A.rules().iterator(); j.hasNext(); ) {
            GLRRule rule = (GLRRule) j.next();
System.out.println("   rule " + rule);

            for (Iterator i = rule.rhs.iterator(); i.hasNext(); ) {
                GLRSymbol X = (GLRSymbol) i.next();

                if (X instanceof GLRNonterminal) {
                    GLRNonterminal B = (GLRNonterminal) X;

                    if (preds[B.index()] == null) {
                        preds[B.index()] = new ArrayList();
                    }

                    preds[B.index()].add(A);

                    collectPostorderDFS(B, postorder, preds, visited);
                }
            }
        }

        postorder.add(A);
    }

    // Partition the recursive nonterminals into sets of mutually
    // recursive nonterminals.
    // This is done by finding SCCs for the graph of nonterminals
    // where there is an edge from A to B if B appears on the RHS
    // of a rule for A.
    void findPartitions(GLRNonterminal S) {
        // First collect all the nonterminals in post order.
        List postorder = new ArrayList();
        List[] preds = new List[nonterminals.size()];
        collectPostorderDFS(S, postorder, preds, new BitSet());

        System.out.println("postorder = " + postorder);

        // Now, walk the transposed graph picking nodes in reverse
        // postorder, thus picking out one SCC at a time and
        // appending it to "by_scc".
        int n = postorder.size();
        BitSet visited = new BitSet();

        Iterator[] p = new Iterator[nonterminals.size()];

	for (int i = 0; i < preds.length; i++) {
            if (preds[i] != null) {
                p[i] = preds[i].iterator();
            }
            else {
                p[i] = Collections.EMPTY_LIST.iterator();
            }
        }

        // Go in reverse postorder
	for (int i = n-1; i >= 0; i--) {
            GLRNonterminal A = (GLRNonterminal) postorder.get(i);

	    if (! visited.get(A.index())) {
		visited.set(A.index());

                GLRNonterminal[] stack = new GLRNonterminal[n];
                int top = 0;

                stack[top++] = A;

                while (top > 0) {
                    GLRNonterminal C = stack[top-1];

                    if (p[C.index()].hasNext()) {
                        GLRNonterminal B = (GLRNonterminal) p[C.index()].next();

                        if (! visited.get(B.index())) {
                            stack[top++] = B;
                        }
                    }
                    else {
                        top--;
                        if (C != A) {
                            System.out.println("union " + C + " and " + A);
                            partitions.union(C.index(), A.index());
                        }
                    }
                }
	    }
	}

	for (int i = 0; i < n; i++) {
            GLRNonterminal A = (GLRNonterminal) postorder.get(i);

            int Ni = partitions.find(A.index());

            if (partition[Ni] == null) {
                partition[Ni] = new ArrayList();
            }

            partition[Ni].add(A);
        }

	for (int i = 0; i < n; i++) {
            GLRNonterminal A = (GLRNonterminal) postorder.get(i);
            int Ni = partitions.find(A.index());
            if (partition[Ni].size() > 1) {
                recursive.set(A.index());
            }
            else if (preds[A.index()] != null) {
                // Check if there's a rule A ->* A
                for (Iterator j = preds[A.index()].iterator(); j.hasNext(); ) {
                    GLRNonterminal B = (GLRNonterminal) j.next();
                    if (A == B) {
                        recursive.set(A.index());
                    }
                }
            }
        }
    }

    // Determine which partitions are left generating and which are right
    // generating.
    void computeLeftRightGeneratingDFS(GLRNonterminal A, BitSet visited) {
        if (visited.get(A.index())) {
            return;
        }

        visited.set(A.index());

        int a = partitions.find(A.index());

        for (Iterator j = A.rules().iterator(); j.hasNext(); ) {
            GLRRule rule = (GLRRule) j.next();

            boolean first = true;

            for (Iterator i = rule.rhs.iterator(); i.hasNext(); ) {
                GLRSymbol X = (GLRSymbol) i.next();

                boolean last = ! i.hasNext();

                if (X instanceof GLRNonterminal) {
                    GLRNonterminal B = (GLRNonterminal) X;
                    computeLeftRightGeneratingDFS(B, visited);

                    int b = partitions.find(B.index());

                    if (a == b) {
                        // Important: use "if", not "else if".  A
                        // nonterminal can be both left and right
                        // generating if the rule is a unit production
                        // A -> B
                        if (! first) {
                            leftGenerating.set(a);
                        }
                        if (! last) {
                            rightGenerating.set(a);
                        }
                    }
                }

                first = false;
            }
        }
    }

    boolean contains(int Ni, GLRSymbol X) {
        if (X instanceof GLRNonterminal) {
            int x = partitions.find(X.index());
            return x == Ni;
        }
        return false;
    }

    boolean containsAny(int Ni, List rhs) {
        for (int i = 0; i < rhs.size(); i++) {
            GLRSymbol X = (GLRSymbol) rhs.get(i);
            if (contains(Ni, X)) {
                return true;
            }
        }
        return false;
    }

    int recursive(int Ni) {
        if (recursive.get(Ni)) {
            if (! leftGenerating.get(Ni) && rightGenerating.get(Ni)) {
                return LEFT;
            }
            if (leftGenerating.get(Ni) && ! rightGenerating.get(Ni)) {
                return RIGHT;
            }
            if (leftGenerating.get(Ni) && rightGenerating.get(Ni)) {
                return SELF;
            }
            if (! leftGenerating.get(Ni) && ! rightGenerating.get(Ni)) {
                return CYCLIC;
            }
        }
        return NONE;
    }

    void makeFAsForList(List alpha) {
        System.out.println("makeFAsForList: " + alpha);

        for (Iterator i = alpha.iterator(); i.hasNext(); ) {
            GLRSymbol X = (GLRSymbol) i.next();

            if (X instanceof GLRNonterminal) {
                GLRNonterminal A = (GLRNonterminal) X;

                if (nfas[A.index()] != null) {
                    continue;
                }

                int Ni = partitions.find(A.index());

                if (recursive(Ni) != SELF) {
                    NonterminalNFA Anfa = new NonterminalNFA(A, alphabet());
                    nfas[A.index()] = Anfa;
    System.out.println("new " + Anfa);

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFA(Anfa, Anfa.q0, rule.rhs, Anfa.q1);
                    }

                    // Now create a DFA for the NFA.
                    DFA Adfa = new DFA(Anfa);
                    nfas[A.index()] = Adfa;
                }
                else {
                    SelfRecursiveApproximationFA pnfa =
                        new SelfRecursiveApproximationFA(alphabet(), nonterminals, this, partition[Ni]);
                    DFA pdfa = new DFA(pnfa, pnfa.startStates());
                    pdfa.lazyBuild();

                    for (Iterator j = partition[Ni].iterator(); j.hasNext(); ) {
                        GLRNonterminal B = (GLRNonterminal) j.next();
                        NFAState nfaStart = pnfa.startStates[B.index()];
                        DFA.DFAState dfaStart = pdfa.startStates[nfaStart.index()];
                        nfas[B.index()] = new WrapperDFA(pdfa, dfaStart);
                    }

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFAsForList(rule.rhs);
                    }
                }
            }
        }
    }

    void makeFA(NonterminalNFA nfa, NFAState q0, List alpha, NFAState q1) {
        System.out.println("makeFA: nfa " + nfa + ", " + alpha);

        if (alpha.isEmpty()) {
            nfa.epsilon(q0, q1);
        }
        else if (alpha.size() > 1) {
            GLRSymbol X = (GLRSymbol) leftHead(alpha);
            NFAState q = nfa.freshState("after-" + X.toString());
            System.out.println("  recurse on " + X);
            makeFA(nfa, q0, Collections.singletonList(X), q);
            System.out.println("  back");
            System.out.println("  recurse on " + leftTail(alpha));
            makeFA(nfa, q, leftTail(alpha), q1);
            System.out.println("  back");
        }
        else if (leftHead(alpha) instanceof GLRTerminal) {
            GLRTerminal a = (GLRTerminal) leftHead(alpha);
            nfa.transition(q0, a, q1);
        }
        else {
            GLRNonterminal A = (GLRNonterminal) leftHead(alpha);

            // building an NFA for A in Nj
            int Nj = partitions.find(nfa.A.index());

            // rhs contains nonterminal for B in Ni
            int Ni = partitions.find(A.index());

            NFAState[] ntStates = new NFAState[nonterminals.size()];

            switch (recursive(Ni)) {
                case NONE: System.out.println("non-recursive"); break;
                case LEFT: System.out.println("left"); break;
                case RIGHT: System.out.println("right"); break;
                case CYCLIC: System.out.println("cyclic"); break;
                case SELF: System.out.println("self"); break;
            }

            if (nfas[A.index()] == null) {
                if (recursive(Ni) != SELF) {
                    NonterminalNFA Anfa = new NonterminalNFA(A, alphabet());
                    nfas[A.index()] = Anfa;
    System.out.println("new " + Anfa);

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFA(Anfa, Anfa.q0, rule.rhs, Anfa.q1);
                    }

                    // Now create a DFA for the NFA.
                    DFA Adfa = new DFA(Anfa);
                    nfas[A.index()] = Adfa;
                }
                else {
                    SelfRecursiveApproximationFA pnfa =
                        new SelfRecursiveApproximationFA(alphabet(), nonterminals, this, partition[Ni]);
                    DFA pdfa = new DFA(pnfa, pnfa.startStates());
                    pdfa.lazyBuild();

                    for (Iterator j = partition[Ni].iterator(); j.hasNext(); ) {
                        GLRNonterminal B = (GLRNonterminal) j.next();
                        NFAState nfaStart = pnfa.startStates[B.index()];
                        DFA.DFAState dfaStart = pdfa.startStates[nfaStart.index()];
                        nfas[B.index()] = new WrapperDFA(pdfa, dfaStart);
                    }

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFAsForList(rule.rhs);
                    }
                }
            }

            switch (recursive(Ni)) {
                case LEFT: {
                    if (Ni != Nj) {
                        nfa.transition(q0, ntTerminal(A), q1);
                    }
                    else {
                        for (Iterator i = partition[Ni].iterator(); i.hasNext(); ) {
                            GLRNonterminal B = (GLRNonterminal) i.next();
                            ntStates[B.index()] = nfa.freshState(B);
                        }

                        NFAState qA = ntStates[A.index()];

                        for (Iterator i = partition[Ni].iterator(); i.hasNext(); ) {
                            GLRNonterminal C = (GLRNonterminal) i.next();

                            NFAState qC = ntStates[C.index()];

                            for (Iterator j = C.rules.iterator(); j.hasNext(); ) {
                                GLRRule rule = (GLRRule) j.next();

                                if (! containsAny(Ni, rule.rhs)) {
                                    makeFA(nfa, q0, rule.rhs, qC);
                                }

                                GLRSymbol D = leftHead(rule.rhs);

                                if (D instanceof GLRNonterminal &&
                                    contains(Ni, D) &&
                                    ! containsAny(Ni, leftTail(rule.rhs))) {
                                    NFAState qD = ntStates[D.index()];
                                    makeFA(nfa, qD, leftTail(rule.rhs), qC);
                                }
                            }
                        }

                        nfa.epsilon(qA, q1);
                    }

                    break;
                }
                case RIGHT:
                case CYCLIC: {
                    if (Ni != Nj && recursive(Ni) != NONE) {
                        nfa.transition(q0, ntTerminal(A), q1);
                    }
                    else {
                        for (Iterator i = partition[Ni].iterator(); i.hasNext(); ) {
                            GLRNonterminal B = (GLRNonterminal) i.next();
                            ntStates[B.index()] = nfa.freshState(B);
                        }

                        NFAState qA = ntStates[A.index()];

                        for (Iterator i = partition[Ni].iterator(); i.hasNext(); ) {
                            GLRNonterminal C = (GLRNonterminal) i.next();

                            NFAState qC = ntStates[C.index()];

                            for (Iterator j = C.rules.iterator(); j.hasNext(); ) {
                                GLRRule rule = (GLRRule) j.next();

                                if (! containsAny(Ni, rule.rhs)) {
                                    makeFA(nfa, qC, rule.rhs, q1);
                                }

                                GLRSymbol D = rightHead(rule.rhs);

                                if (D instanceof GLRNonterminal &&
                                    contains(Ni, D) &&
                                    ! containsAny(Ni, rightTail(rule.rhs))) {
                                    NFAState qD = ntStates[D.index()];
                                    makeFA(nfa, qC, rightTail(rule.rhs), qD);
                                }
                            }
                        }

                        nfa.epsilon(q0, qA);
                    }

                    break;
                }
                case SELF: {
                    nfa.transition(q0, ntTerminal(A), q1);

                    // Make sure DFAs are build for all the nonterminals
                    // reachable from A.
                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFAsForList(rule.rhs);
                    }

                    break;
                }
                default: {
                    // not recursive
                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFA(nfa, q0, rule.rhs, q1);
                    }
                    break;
                }
            }
        }
    }

    static class NTTerminal implements NFATerminal {
        GLRNonterminal A;
        int index;
        NTTerminal(GLRNonterminal A, int index) {
            this.A = A;
            this.index = index;
        }
        public int index() { return index; }
        public String toString() {
            return "t" + A.toString();
        }
    }

    NFATerminal ntTerminal(GLRNonterminal A) {
        NFATerminal a = ntTerminals[A.index()];
        if (a == null) {
            a = new NTTerminal(A, alphabet.size());
            ntTerminals[A.index()] = a;
            alphabet.add(a);
        }
        return a;
    }

    NFA dfaForSymbol(GLRSymbol X) {
        System.out.println("getting DFA for " + X);

        if (X instanceof GLRNonterminal) {
            GLRNonterminal A = (GLRNonterminal) X;
            NFA dfa = nfas[A.index()];
            if (dfa instanceof DFA) {
                ((DFA) dfa).lazyBuild();
            }
            return dfa;
        }
        else {
            return new TerminalDFA((GLRTerminal) X, alphabet);
        }
    }

    static class AcceptDFA implements NFA {
        List alphabet;
        NFAState startState;

        public AcceptDFA(List alphabet) {
            this.alphabet = alphabet;
            this.startState = new NFAState(0) {
                public String toString() { return "accept"; } };
        }

        public boolean equals(Object o) {
            if (o instanceof AcceptDFA) {
                AcceptDFA m = (AcceptDFA) o;
                return true;
            }
            return false;
        }

        public String toString() {
            return "accept DFA";
        }

        public int size() {
            return 1;
        }

        public Iterator succs(Node node) {
            return Collections.EMPTY_LIST.iterator();
        }

        public Iterator preds(Node node) {
            return Collections.EMPTY_LIST.iterator();
        }

        public Iterator nodes() {
            return Collections.singleton(startState).iterator();
        }

        public boolean accepting(NFAState q0, NFAState q) {
            return q0 == startState && q == startState;
        }

        public NFAState nodeAt(int i) {
            if (i == 0) return startState;
            return null;
        }

        public List alphabet() {
            return alphabet;
        }

        public NFAState start() {
            return startState;
        }

        public void dump() {
            System.out.println(toString());
        }
    }

    static class TerminalDFA implements NFA {
        GLRTerminal label;
        List alphabet;
        NFAState startState;
        NFAState acceptState;

        public TerminalDFA(final GLRTerminal label, List alphabet) {
            this.label = label;
            this.alphabet = alphabet;
            this.startState = new NFAState(0) {
                public String toString() { return label + "0"; } };
            this.acceptState = new NFAState(1) {
                public String toString() { return label + "1"; } };
        }

        public boolean equals(Object o) {
            if (o instanceof TerminalDFA) {
                TerminalDFA m = (TerminalDFA) o;
                return label.equals(m.label);
            }
            return false;
        }

        public String toString() {
            return "DFA for " + label;
        }

        public int size() {
            return 2;
        }

        public Iterator succs(Node node) {
            if (node == startState) {
                NFAEdge edge = new NFAEdge(startState, acceptState, label);
                return Collections.singleton(edge).iterator();
            }
            return Collections.EMPTY_LIST.iterator();
        }

        public Iterator preds(Node node) {
            if (node == acceptState) {
                NFAEdge edge = new NFAEdge(acceptState, startState, label);
                return Collections.singleton(edge).iterator();
            }
            return Collections.EMPTY_LIST.iterator();
        }

        public Iterator nodes() {
            List l = new ArrayList(2);
            l.add(startState);
            l.add(acceptState);
            return l.iterator();
        }

        public NFAState nodeAt(int i) {
            if (i == 0) return startState;
            if (i == 1) return acceptState;
            return null;
        }

        public List alphabet() {
            return alphabet;
        }

        public boolean accepting(NFAState q0, NFAState q) {
            return q0 == startState && q == acceptState;
        }

        public NFAState start() {
            return startState;
        }

        public void dump() {
            System.out.println(toString());
        }
    }

    static class NFAList {
        NFA head;
        NFAList tail;

        NFAList(NFA head, NFAList tail) {
            this.head = head;
            this.tail = tail;
        }

        public boolean equals(Object o) {
            if (o instanceof NFAList) {
                NFAList l = (NFAList) o;
                if (head.equals(l.head)) {
                    if (tail == null) return l.tail == null;
                    return tail.equals(l.tail);
                }
            }
            return false;
        }
    }

    NFAList forList(List alpha) {
        if (alpha.isEmpty()) {
            return null;
        }
        else {
            GLRSymbol X = leftHead(alpha);
            NFA dfa = dfaForSymbol(X);
            return new NFAList(dfa, forList(leftTail(alpha)));
        }
    }

    /**
     * Return true if the language of symbols <code>alpha1</code> intersects
     * the language of symbols <code>alpha2</code>.
     */
    public boolean intersects(List alpha1, List alpha2) {
        NFAList dfas1 = forList(alpha1);
        NFAList dfas2 = forList(alpha2);

        if (dfas1 == null) {
            dfas1 = new NFAList(new AcceptDFA(alphabet), null);
        }

        if (dfas2 == null) {
            dfas2 = new NFAList(new AcceptDFA(alphabet), null);
        }

        NFAState s1 = dfas1.head.start();
        NFAState s2 = dfas2.head.start();

        return intersectsDFS(dfas1, dfas2, s1, s2, new HashSet());
    }

    static class Key {
        NFAList l1, l2;
        NFAState s1, s2;

        Key(NFAList l1, NFAList l2, NFAState s1, NFAState s2) {
            this.l1 = l1;
            this.l2 = l2;
            this.s1 = s1;
            this.s2 = s2;
        }

        public int hashCode() {
            return s1.hashCode() + s2.hashCode();
        }

        public boolean equals(Object o) {
            if (o instanceof Key) {
                Key k = (Key) o;
                return l1.equals(k.l1) && l2.equals(k.l2)
                    && s1.equals(k.s1) && s2.equals(k.s2);
            }
            return false;
        }
    }

    public boolean intersectsDFS(NFAList dfas1, NFAList dfas2,
                                 NFAState s1, NFAState s2, Set visited)
    {
        Key k = new Key(dfas1, dfas2, s1, s2);

        if (visited.contains(k)) {
            return false;
        }

        visited.add(k);

        NFA dfa1 = dfas1.head;
        NFA dfa2 = dfas2.head;

        if (Report.should_report("approx", 4)) {
            Report.report(4, "intersectDFS: " + s1 + " in " + dfa1 +
                                      " and " + s2 + " in " + dfa2);
        }

        boolean accept1 = false;
        boolean accept2 = false;

        while (dfa1.accepting(dfa1.start(), s1)) {
            if (Report.should_report("approx", 4)) {
                Report.report(4, "intersectDFS: accept left in " + s1);
            }

            if (dfas1.tail != null) {
                dfas1 = dfas1.tail;
                dfa1 = dfas1.head;
                s1 = dfa1.start();
            }
            else {
                accept1 = true;
                break;
            }
        }

        while (dfa2.accepting(dfa2.start(), s2)) {
            if (Report.should_report("approx", 4)) {
                Report.report(4, "intersectDFS: accept right in " + s2);
            }

            if (dfas2.tail != null) {
                dfas2 = dfas2.tail;
                dfa2 = dfas2.head;
                s2 = dfa2.start();
            }
            else {
                accept2 = true;
                break;
            }
        }

        if (Report.should_report("approx", 4)) {
            Report.report(4, "intersectDFS: " + s1 + " in " + dfa1 +
                                      " and " + s2 + " in " + dfa2);
        }

        if (accept1 && accept2) {
            if (Report.should_report("approx", 4)) {
                Report.report(4, "intersectDFS: accept!");
            }
            return true;
        }

        // Get the labels in common.
        Map labels1 = new HashMap();
        Map labels2 = new HashMap();

        for (Iterator i = dfa1.succs(s1); i.hasNext(); ) {
            NFAEdge e = (NFAEdge) i.next();

            if (e.label instanceof NTTerminal) {
                NTTerminal t = (NTTerminal) e.label;
                NFA dfa = dfaForSymbol(t.A);
                NFAState s = dfa.start();
                NFAList l = new NFAList(dfa, dfas1.tail);

                if (Report.should_report("approx", 4)) {
                    Report.report(4, "intersectDFS: recursing left on " + s);
                }

                if (intersectsDFS(l, dfas2, s, s2, visited)) {
                    if (Report.should_report("approx", 4)) {
                        Report.report(4, "intersectDFS: accept after " + s + " and " + s2);
                    }
                    return true;
                }
            }
            else {
                labels1.put(e.label, e);
            }
        }

        for (Iterator i = dfa2.succs(s2); i.hasNext(); ) {
            NFAEdge e = (NFAEdge) i.next();

            if (e.label instanceof NTTerminal) {
                NTTerminal t = (NTTerminal) e.label;
                NFA dfa = dfaForSymbol(t.A);
                NFAState s = dfa.start();
                NFAList l = new NFAList(dfa, dfas2.tail);

                if (Report.should_report("approx", 4)) {
                    Report.report(4, "intersectDFS: recursing right on " + s);
                }

                if (intersectsDFS(dfas1, l, s1, s, visited)) {
                    if (Report.should_report("approx", 4)) {
                        Report.report(4, "intersectDFS: accept after " + s1 + " and " + s);
                    }
                    return true;
                }
            }
            else {
                labels2.put(e.label, e);
            }
        }

        // Intersect the edges.
        for (Iterator i = labels1.keySet().iterator(); i.hasNext(); ) {
            NFATerminal a = (NFATerminal) i.next();
            Edge e1 = (Edge) labels1.get(a);
            Edge e2 = (Edge) labels2.get(a);

            if (e1 == null) continue;
            if (e2 == null) continue;

            if (Report.should_report("approx", 4)) {
                Report.report(4, "intersectDFS: common label " + a);
            }

            NFAState t1 = (NFAState) e1.target();
            NFAState t2 = (NFAState) e2.target();

            if (intersectsDFS(dfas1, dfas2, t1, t2, visited)) {
                if (Report.should_report("approx", 4)) {
                    Report.report(4, "intersectDFS: accept after " + t1 + " and " + t2);
                }
                return true;
            }
        }

        if (Report.should_report("approx", 4)) {
            Report.report(4, "intersectDFS: no common labels at " + s1 + " and " + s2);
        }

        return false;
    }
}
