package polyglot.ext.ibex.lr;

import polyglot.ext.ibex.types.*;
import polyglot.util.*;
import java.util.*;
import polyglot.main.Report;

public class GLR {
    List nonterminals;
    List terminals;
    List rules;
    GLRTerminal eofSymbol;
    List startSymbols;

    LRConstruction lrc;
    LREncoding lre;
    MergeDetection md;

    /**
     * Map from source grammar start symbols to augmented grammar start
     * symbols.
     */
    GLRNonterminal[] startSymbolsMap;
    List augmentedStartSymbols;

    /** Map from type system symbols to LR symbols. */
    Map symbolMap;

    public GLR(ParserType pt) {
        initSymbols(pt);
        initRules(pt);

        augmentGrammar();

        computeNullable();
        computeFirst();
        computeFollow();

        if (Report.should_report(TOPICS, 9) ||
            Report.should_report("dump-lr", 1)) {
            dump();
        }

        lrc = new LRConstruction(augmentedStartSymbols, terminals, nonterminals, rules, eofSymbol());

        if (Report.should_report(TOPICS, 9) ||
            Report.should_report("dump-lr", 1)) {
            dump();
        }
    }

    // -------------------- getters and setters --------------------
    List nonterminals() { return nonterminals; }
    List rules() { return rules; }
    GLRTerminal eofSymbol() { return eofSymbol; }

    public int numRules() {
        return rules.size();
    }

    public Nonterminal ruleLhs(int i) {
        GLRRule rule = (GLRRule) rules.get(i);
        return rule.nonterm;
    }

    public RhsCase ruleRhs(int i) {
        GLRRule rule = (GLRRule) rules.get(i);
        return rule.nontermRhs;
    }

    public int eofSymbolNumber() {
        return eofSymbol().index();
    }

    public int ruleNumber(Nonterminal lhs, RhsCase rhs) {
        GLRNonterminal t = nonterminal(lhs);

        for (Iterator i = t.rules().iterator(); i.hasNext(); ) {
            GLRRule r = (GLRRule) i.next();
            if (r.nontermRhs.hasSameElements(rhs)) {
                return r.index();
            }
        }

        throw new InternalCompilerError("Could not get rule number for " + lhs + " ::= " + rhs);
    }

    public int nonterminalNumber(Nonterminal s) {
        GLRNonterminal t = nonterminal(s);
        return t.index();
    }

    public int terminalNumber(Terminal s) {
        GLRTerminal t = terminal(s);
        return t.index();
    }

    public boolean isStartSymbol(Nonterminal s) {
        GLRNonterminal t = nonterminal(s);
        return lrc.startStatesMap[t.index()] != null;
    }

    /**
     * Get the start symbol number for a particular nonterminal <code>s</code>.
     * This should only be called if the symbol <code>s</code> is a start
     * symbol.  This method is needed to generate code to invoke the parser for
     * a particular symbol.
     */
    public int startSymbolNumber(Nonterminal s) {
        GLRNonterminal t = nonterminal(s);
        t = startSymbolsMap[t.index()];
        if (t == null) {
            throw new InternalCompilerError("Symbol " + s +
                                            " is not a start symbol.");
        }
        return t.index();
    }

    /**
     * Get the start state number for a particular nonterminal <code>s</code>.
     * This should only be called if the symbol <code>s</code> is a start
     * symbol.  This method is needed to generate code to invoke the parser for
     * a particular symbol.
     */
    public int startStateNumber(Nonterminal s) {
        State I = lrc.startStatesMap[startSymbolNumber(s)];
        if (I == null) {
            throw new InternalCompilerError("Symbol " + s +
                                            " is not a start symbol.");
        }
        return I.index();
    }

    public String[] encodedActionTable() {
        if (lre == null) {
            lre = new LREncoding(this);
        }
        return lre.encodedActionTable();
    }

    public String[] encodedOverflowTable() {
        if (lre == null) {
            lre = new LREncoding(this);
        }
        return lre.encodedOverflowTable();
    }

    public String[] encodedGotoTable() {
        if (lre == null) {
            lre = new LREncoding(this);
        }
        return lre.encodedGotoTable();
    }

    public String[] encodedRuleTable() {
        if (lre == null) {
            lre = new LREncoding(this);
        }
        return lre.encodedRuleTable();
    }

    /**
     * Return a collection of collections of right-hand-sides
     * <code>rhs</code> associated with the nonterminal <code>lhs</code>
     * such that the rule <code>lhs ::= rhs</code> requires a merge
     * action.  Each pair of rules in each collection should have a merge
     * action.
     */
    public Collection mergeRulesForNonterminal(Nonterminal lhs) {
        if (md == null) {
            md = new MergeDetection(this);
        }
        GLRNonterminal n = nonterminal(lhs);
        return md.mergeRulesForNonterminal(n);
    }

    /** Initialize the symbols of the grammar. */
    void initSymbols(ParserType pt) {
        symbolMap = new HashMap();
        nonterminals = new ArrayList(pt.allNonterminals().size()+1);
        terminals = new ArrayList(pt.allTerminals().size()+1);
        startSymbols = new ArrayList();

        // Force symbol numbering to be the same as the source order
        // by creating the symbols now.
        for (Iterator i = pt.allTerminals().iterator(); i.hasNext(); ) {
            Terminal r = (Terminal) i.next();
            terminal(r);
        }

        for (Iterator i = pt.allNonterminals().iterator(); i.hasNext(); ) {
            Nonterminal r = (Nonterminal) i.next();
            GLRNonterminal n = nonterminal(r);
            if (r.isStart()) {
                startSymbols.add(n);
            }
        }

        if (nonterminals.isEmpty() || terminals.isEmpty()) {
            throw new InternalCompilerError(pt + " is not a parser.");
        }

        if (startSymbols.isEmpty()) {
            throw new InternalCompilerError(pt + " has no start symbols.");
        }

        if (eofSymbol == null) {
            eofSymbol = (GLRTerminal) new GLRTerminal("$", terminals.size());
            terminals.add(eofSymbol);
        }
    }

    /** Initialize the rules of the grammar. */
    void initRules(ParserType pt) {
        rules = new ArrayList(pt.allNonterminals().size()*2);

        for (Iterator i = pt.allNonterminals().iterator(); i.hasNext(); ) {
            Nonterminal r = (Nonterminal) i.next();

            GLRNonterminal n = nonterminal(r);

            for (Iterator j = r.choices().iterator(); j.hasNext(); ) {
                Rhs choice = (Rhs) j.next();

                if (! (choice instanceof RhsCase)) {
                    continue;
                }

                RhsCase rhs = (RhsCase) choice;

                List l = new ArrayList(rhs.elements().size());

                for (Iterator k = rhs.elements().iterator(); k.hasNext(); ) {
                    Symbol e = (Symbol) k.next();

                    if (e instanceof Terminal) {
                        Terminal s = (Terminal) e;
                        l.add(terminal(s));
                    }
                    else if (e instanceof Nonterminal) {
                        Nonterminal s = (Nonterminal) e;
                        GLRNonterminal t = nonterminal(s);
                        l.add(t);
                    }
                }

                GLRRule rule = new GLRRule(r, rhs, n, l, rules.size());
                rules.add(rule);

                n.rules().add(rule);
            }
        }

        // To avoid the yield-before-merge problem,
        // the parser compares rules to be reduced so
        // that if B ->+ A, a reduction of A occurs before
        // a reduction of B.  To make this comparison simpler
        // we generate the rules so that if B ->+ A, A occurs
        // first in the rules list.
        //
        // Sort the rules by the derives relation.
        sortRules();
    }

    // Sort the rules by the derives relation.
    void sortRules() {
        Collections.sort(rules, new Comparator() {
            public int compare(Object o1, Object o2) {
                GLRRule r1 = (GLRRule) o1;
                GLRRule r2 = (GLRRule) o2;
                return r1.index - r2.index; // FIXME: should sort by
                                            // derives relation
            }
        });

        // Now, renumber the rules.
        for (int i = 0; i < rules.size(); i++) {
            GLRRule r = (GLRRule) rules.get(i);
            r.index = i;
        }
    }

    /** Augment the grammar with start rules for the start symbols. */
    void augmentGrammar() {
        startSymbolsMap = new GLRNonterminal[nonterminals.size() +
                                            startSymbols.size()];
        augmentedStartSymbols = new ArrayList(startSymbols.size());

        // For each start symbol S, create a new rule S' ::= S $
        for (Iterator i = startSymbols.iterator(); i.hasNext(); ) {
            GLRNonterminal s = (GLRNonterminal) i.next();

            GLRNonterminal s_ = new GLRNonterminal(s.nonterm,
                                                  s.name() + "!",
                                                  nonterminals.size());
            nonterminals.add(s_);

            // create a new rule for the new start symbol, S'
            // S' ::= S $
            List l = new ArrayList(2);
            l.add(s);
            l.add(eofSymbol());

            GLRRule r = new GLRRule(null, null, s_, l, rules.size());
            rules.add(r);
            s_.rules.add(r);

            startSymbolsMap[s.index()] = s_;
            augmentedStartSymbols.add(s_);
        }
    }

    void dump() {
        for (int i = 0; i < terminals.size(); i++) {
            GLRTerminal t = (GLRTerminal) terminals.get(i);
            System.out.println("Terminal " + i + ": " + t);
        }
        System.out.println();

        for (int i = 0; i < nonterminals.size(); i++) {
            GLRNonterminal t = (GLRNonterminal) nonterminals.get(i);
            System.out.println("Nonterminal " + i + ": " + t);
        }
        System.out.println();

        for (int i = 0; i < rules.size(); i++) {
            GLRRule t = (GLRRule) rules.get(i);
            System.out.println("Rule " + t);
        }
        System.out.println();

        if (lrc != null)
            lrc.dump();
    }

    GLRNonterminal nonterminal(Nonterminal n) {
        GLRNonterminal s = (GLRNonterminal) symbolMap.get(n);

        if (s == null) {
            s = new GLRNonterminal(n, n.name(), nonterminals.size());
            nonterminals.add(s);
            symbolMap.put(n, s);
        }

        return s;
    }

    GLRTerminal terminal(Terminal n) {
        GLRTerminal s = (GLRTerminal) symbolMap.get(n);

        if (s == null) {
            s = new GLRTerminal(n.name(), terminals.size());
            terminals.add(s);
            symbolMap.put(n, s);
        }

        return s;
    }

    // Compute nullable(X) for all nonterminals, X.
    private void computeNullable() {
        // Assume nullable flags initialized to false.

        if (Report.should_report(TOPICS, 1))
            Report.report(1, "computing nullable sets");

        boolean changed = true;

        while (changed) {
            changed = false;

            for (Iterator it = rules().iterator(); it.hasNext(); ) {
                GLRRule rule = (GLRRule) it.next();

                GLRNonterminal lhs = rule.lhs();
                List rhs = rule.rhs();

                boolean allNullable = true;

                for (Iterator rit = rhs.iterator(); rit.hasNext(); ) {
                    GLRSymbol s = (GLRSymbol) rit.next();

                    if (s instanceof GLRNonterminal) {
                        GLRNonterminal nt = (GLRNonterminal) s;

                        if (! nt.isNullable()) {
                            allNullable = false;
                            break;
                        }
                    }
                    else {
                        allNullable = false;
                        break;
                    }
                }

                if (allNullable && ! lhs.isNullable()) {
                    changed = true;
                    lhs.setNullable(true);
                }
            }
        }
    }

    // Compute first(X) for all nonterminals, X.
    private void computeFirst() {
        // Assume first sets initialized to empty.

        if (Report.should_report(TOPICS, 1))
            Report.report(1, "computing first sets");

        boolean changed = true;

        while (changed) {
            changed = false;

            for (Iterator it = rules().iterator(); it.hasNext(); ) {
                GLRRule rule = (GLRRule) it.next();

                GLRNonterminal lhs = rule.lhs();
                List rhs = rule.rhs();

                Set firstRHS = new HashSet();

                // Iterate over the RHS = Y1 Y2 ... Yk and add in FIRST(Yi)
                // until Yi is not nullable.
                for (Iterator rit = rhs.iterator(); rit.hasNext(); ) {
                    GLRSymbol s = (GLRSymbol) rit.next();

                    if (s instanceof GLRNonterminal) {
                        GLRNonterminal nt = (GLRNonterminal) s;

                        firstRHS.addAll(nt.first());

                        if (! nt.isNullable()) {
                            break;
                        }
                    }
                    else if (s instanceof GLRTerminal) {
                        firstRHS.add(s);
                        break;
                    }
                }

                // Union FIRST(rhs) into FIRST(lhs) and see if the set grew.
                changed |= lhs.first().addAll(firstRHS);
            }
        }
    }

    // Compute follow(X) for all nonterminals, X.
    private void computeFollow() {
	// Assume first sets initialized to empty.

        // unnecessary, since in the rules
        // startSymbol().follow().add(eofSymbol());

	boolean changed = true;

	while (changed) {
	    changed = false;

	    for (Iterator it = rules().iterator(); it.hasNext(); ) {
		GLRRule rule = (GLRRule) it.next();

		GLRNonterminal lhs = rule.lhs();
		LinkedList rhs = new LinkedList(rule.rhs());

		while (! rhs.isEmpty()) {
		    GLRSymbol e = (GLRSymbol) rhs.removeFirst();

		    if (e instanceof GLRNonterminal) {
			GLRNonterminal s = (GLRNonterminal) e;

			// The rule we're examining looks like:
			//     X -> Y1 Y2 ... Yi Yi+1 ... Yk
			// s is Yi and rhs contains Yi+1 ... Yk
			// FOLLOW(Yi) contains FIRST(Yi+1 ... Yk)
			// and also FOLLOW(Yi+1 ... Yi+j) if that is nullable.

			Set followSym = new HashSet();
			boolean allNullable = true;

                        for (Iterator rit = rhs.iterator(); rit.hasNext(); ) {
			    GLRSymbol e2 = (GLRSymbol) rit.next();

			    if (e2 instanceof GLRNonterminal) {
				GLRNonterminal nt = (GLRNonterminal) e2;

				followSym.addAll(nt.first());

				if (! nt.isNullable()) {
				    allNullable = false;
				    break;
				}
			    }
			    else if (e2 instanceof GLRTerminal) {
				GLRTerminal t = (GLRTerminal) e2;
				followSym.add(t);
				allNullable = false;
				break;
			    }
			}

			if (allNullable) {
			    followSym.addAll(lhs.follow());
			}

			changed |= s.follow().addAll(followSym);
		    }
		}
	    }
	}
    }

    static Collection TOPICS = Arrays.asList( new String[] { "lr", "ibex" });
}
