package polyglot.ext.polyj.ast;

import polyglot.ext.polyj.extension.PolyJExt;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.ast.*;

import polyglot.ast.*;
import polyglot.frontend.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import java.util.*;

/** 
 * A <code>PolyJFieldDecl</code> is an immutable
 * representation of a variable declaration, which consists of a type, one or
 * more variable names, and possible initilization expressions.
 */
public class PolyJFieldDecl_c extends FieldDecl_c implements PolyJFieldDecl {

    /**
     * Creates a new <code>PolyJFieldDecl</code>.
     */
    public PolyJFieldDecl_c(Position pos,
                            Flags flags,
                            TypeNode type,
                            String name,
                            Expr init) {
        super( pos, flags, type, name, init);
    }

    public Node typeCheck(TypeChecker tc) throws SemanticException {
        Context c = tc.context();
        ClassType currentClass = c.currentClass();
        PolyJTypeSystem ts = (PolyJTypeSystem) tc.typeSystem();
        if (flags().isStatic() && flags().isPrivate() 
            && ts.needsWhereObject(currentClass)) 
        {
            throw new SemanticException("PolyJ does not allow private static fields"
                + " in parametric classes.");
        }
        
        return super.typeCheck(tc);
    }
    
    public void translate(CodeWriter w, Translator tr) {
	Context c = tr.context();                                                       
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
	
        ClassType currentClass = c.currentClass();
	boolean isParameterized = ts.needsWhereObject(currentClass);

        // Do not put static field into the class --
        // it will go into the instantiation adapter classes
        if (flags().isStatic() && isParameterized) {
            writeStaticInit(w, (PolyJTranslator) tr, c, ts);
        } else {
            super.translate(w, tr);
	}
    }
    
    /**
     * Write the initializer method for the field.
     */
    protected void writeStaticInit(CodeWriter w, PolyJTranslator tr, Context c,
        PolyJTypeSystem ts) 
    {
        ClassType ct = c.currentClass();
        String adIntName = PolyJTranslator.adapterInterfaceShortName(ct);
        w.write("private static void $init$" + name() + "(" + adIntName 
            + " $W) { ");
        
        Node n = getInit(tr, c, ct);
        if (n != null) {       
            w.newline(4);
            w.begin(0);
            n.del().translate(w, tr);
            w.end();
            w.newline(0);    
        }
        
        w.write("}");
        w.newline(0);
    }
    
    /**
     * Return a PolyJ AST that represents the initializer for the
     * field.  This may need to be generated if no initializer was
     * provided in the source, because parametric fields need
     * a default initialization.  Returns null if no initializer
     * is necessary.
     */
    protected Node getInit(PolyJTranslator tr, Context ctxt, ClassType ct) {
        Node n = null;
	if (init() != null) {
	    n = tr.qq().parseStmt(name() + " = %E;", init());
	} else {
	    PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
	    FieldInstance origFi = ts.originalField(ct, fieldInstance());
	    Type origType = origFi.type();
	    if (origType instanceof ParamType) {
		ParamType p = (ParamType) origType;
		n = tr.qq().parseStmt(name() + " = (" + p.name() 
		    + ".default);");
	    }
	}
	if (n != null) {
	    n = tr.job().spawn(ctxt, n, Pass.BUILD_TYPES, Pass.PRE_OUTPUT_ALL).ast();
	}
        return n;
    }
    
    public void dump(CodeWriter w) {
        w.write( "POLYJ FIELD DECL");
        w.write( " < " + flags().toString() + "> ");
    }
}
