package polyglot.ext.polyj.types;

import polyglot.frontend.*;
import polyglot.types.*;
import polyglot.util.*;

import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import java.util.*;

public interface PolyJTypeSystem extends TypeSystem {

    TranslationTable translationTable();
    void debug_dumpTranslationTable();

    /**
     * Return true iff t is a type that should have an
     * adapter class generated for it.
     */
    boolean needsAdapterClass(Type t);

    /**
     * Return whether all the type parameters of <code>t</code>
     * have been instantiated.
     */
    boolean allParamsInst(ClassType t);
    
    /**
     * Return whether t has been fully instantiated on
     * all its type parameters.
     */
    boolean isFullyInst(ClassType t);

    /**
     * Return whether t is a type parameter.
     */
    boolean isParameter(Type t);

    /**
     * Return whether t is a class type that could
     * be or is instantiated.
     */
    boolean isPolymorphicClass(Type t);

    /**
     * Return whether c has been instantiated.
     */
    boolean isInstantiated(ClassType c);

    /**
     * returns true iff rt is an instantiation of c
     */
    boolean isInstantiationOf(InstType rt, PolyJClassType c);

    /**
     * Return true if t1 and t2 represent the same instantiated type.
     */
    boolean isSameInst(ClassType t1, ClassType t2);
    
    /**
     * Return the uninstantiated type of c.
     */
    PolyJClassType uninstantiate(ClassType c);

    /**
     * Return all the actual type arguments used in forming t, in
     * order from the outermost enclosing type to the innermost.
     * For example, if t is C[int].D[List[char]], returns [int, List[char]].
     * The elements of the list are Types.
     * If t has no such arguments, returns the empty list.
     */
    List allTypeArgs(ClassType t);

    /**
     * Return all the formal type parameters used in forming t, in
     * order from the outermost enclosing type to the innermost.
     * For example, if t is C[T].D[U], returns [T,U].
     * The elements of the list are ParamTypes.
     * If t has no such arguments, returns the empty list.
     */
    List allTypeParams(ClassType t);

    /**
     * check whether the given types can be used to satisfy
     * the constraints of the given polymorphic type.
     * @param t the polymorphic type
     * @param types the types to instantiate on
     * @param instantiationContext the place where the instantiation
     *        occurs.  This determines the visibility and applicability
     *        of methods for the purposes of constraint satisfaction.
     * @return true if it would be ok to instantiate t on the given types
     * @exception ConstraintException if t cannot be instantiated on types.
     */
    void checkConstraints(PolyJClassType t, List types,
            Context instantiationContext) throws ConstraintException;

    /**
     * Check that instArgs has the correct number and type of arguments to
     * instantiate instType.
     */
    void checkInstArgs(PolyJClassType instType, List instArgs) throws SemanticException;

    /**
     * Create the type wrapper responsible for instantiating the
     * given polymorphic type t on the argument types.
     */
    PolyJTypeWrapper Wrapper(PolyJClassType t, List types);

    /**
     * Find the method in implementingType that has the
     * signature of paramMethod modulo substitutions done by
     * the wrapper, in the given context.
     */
    MethodInstance correspondingMethod(MethodInstance paramMethod,
                                       ParamType constrainingType,
                                       ClassType implementingType,
                                       PolyJTypeWrapper wrapper,
                                       Context ctxt)
        throws SemanticException;

    /**
     * get the method in the original (completely uninstantiated) type
     * that corresponds to this method instance.
     */
    MethodInstance originalMethod(MethodInstance mi);
    ConstructorInstance originalConstructor(ConstructorInstance ci);
    
    /**
     * Find the original FieldInstance in <code>c</code> that defines
     * field <code>fi</code>.
     */
    FieldInstance originalField(ClassType c, FieldInstance fi);
    
    /**
     * Return whether p1 and p2 have the same arguments, as they
     * will be implemented in the translation.  That is, a
     * parameter could be a primitive or Object, and a polymorphic type
     * will be Object.
     */
    boolean hasSameImplArguments(ProcedureInstance p1, ProcedureInstance p2);

    /**
     * Instantiate t and its outer classes on their own parameters.
     */
    ClassType nullInstantiate(PolyJClassType t);

    InstType instantiate(PolyJClassType t, List args) throws SemanticException;
    InstType instantiate(Position pos, PolyJClassType t, List args) throws SemanticException;
    InstType uncheckedInstantiate(PolyJClassType t, List args);
    InstType uncheckedInstantiate(Position pos, PolyJClassType t, List args);

    /**
     * Record in the translator that c is a type that was
     * instantiated in the source.
     */
    void addInstantiation(ClassType c);
    void addInstantiations(Collection types);

    /**
     * Assisted instantiations are situations where a polymorphic type
     * is instantiating another polymorphic type with one or more of
     * its own parameter type.  We need a method in the where object
     * to return a where object for the instance class.
     */
    void addAssistedInstantiation(PolyJClassType assistant, ClassType instance);

    void addNewArrayMethod(ParamType param, int dims);

    AdapterBinding AdapterBinding(PolyJClassType assistant,
        ClassType instantiation);
    ArrayAccessBinding ArrayAccessBinding(ParamType param,
        ClassType enclosing);
    DefaultBinding DefaultBinding(ParamType param, ClassType enclosing);
    InstantiationBinding InstantiationBinding(ClassType c);
    NewArrayBinding NewArrayBinding(ParamType param, int dims,
        ClassType enclosing);
    ParamSigBinding ParamSigBinding(ParamType param, ClassType enclosing);
    StaticFieldBinding StaticFieldBinding(ClassType enclosingType, 
        FieldInstance fi);
    WhereBinding WhereBinding(ParamType param, ProcedureInstance constraint);

    /**
     * Return whether any of <code>c</code>'s enclosing instances
     * have a where object field.
     */
    boolean hasOuterWhereObject(ClassType c);
    
    /**
     * Returns true iff some class up the type hierarchy from t
     *   has a where-object field.
     *
     * A class inherits a where-object if its super-type needs a
     * where object (which is to say, its super-type has a where object).
     */
    boolean inheritsWhereObject(ReferenceType t);

    /**
     * A class needs a where-object if its an instantiated type or
     * a polymorphic type, or if its supertype or one of its super-interfaces
     * needs a where object
     */
    boolean needsWhereObject(ReferenceType t);

    /**
     * A prmitive type wrapper is the runtime representation of a
     * primitive type when it is manipulated by a polymorphic type.
     */
    String primitiveTypeWrapperName(PrimitiveType primitive);

    /**
     * Proxy types double for non-class types during constraint checking
     * of instantiated types.
     */
    Type instantiationProxyType(PrimitiveType primitive);
    Type instantiationProxyType(ArrayType ary);

    String translateArray(Resolver c, ArrayType array);

    ParamType ParamType(Position pos, PolyJClassType constrainedClass,
            String paramName, int index);
    PolyDlg PolyDlg(PolyJClassType ct);
    WrapDlg WrapDlg(WrappedClassType w, PolyJClassType r,
            ClassType oldEnclosingType, ClassType newEnclosingType,
            PolyJTypeWrapper tw);
    WrappedClassType WrappedClassType(PolyJClassType r, ClassType oldEnclosingType,
            ClassType newEnclosingType, PolyJTypeWrapper tw);
}
